/*
Code is partly based on pseudo code out of the paper:
"There Is No 16-Clue Sudoku: Solving the Sudoku Minimum Number of Clues Problem via Hitting Set Enumeration".
It was written by Gary McGuire, Bastian Tugemann, and Gilles Civario in 2014 and
published in Experimental Mathematics, 23:190217
*/

#include <iostream>
#include <vector>
#include <cmath>
#include <algorithm>
#include <chrono>
#include <string>
#include <memory>

#include <iomanip>
#include <queue>
#include <stdint.h>

#include<cstdlib>


#include <boost/dynamic_bitset.hpp>

#include "Blueprints.h"
#include "SudokuGrid.h"
#include "getRSS.c"

#include <stdint.h>

#include <fstream>

//  Windows
#ifdef _WIN32

#include <intrin.h>

uint64_t rdtsc(){
    return __rdtsc();
}

//  Linux/GCC
#else

uint64_t rdtsc(){
    unsigned int lo,hi;
    __asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi));
    return ((uint64_t)hi << 32) | lo;
}
#endif

using namespace std;
typedef unsigned char byte;

int maxClues = 16;

// some debug variables
vector<int> debugAssignVec(1536);
vector<byte> debugMissedVec{0, 58, 32, 71, 27, 23, 74, 59, 52, 47, 12, 26, 78};
vector<byte> debugMissedVec2{ 0, 58, 32, 71, 27, 23 };
bool debugbool = false;

//for counting
clock_t timeSpentCheckingCandidates = 0;
uint64_t countHittingSets = 0;
uint64_t countSmallerHittingSets = 0;
uint64_t countLastLayerNotAllHit = 0;
vector<uint64_t> countIgnored(maxClues, 0);
vector<uint64_t> countNotIgnored(maxClues, 0);
vector<uint64_t> countDegreeBacktracked(6);
vector<uint64_t> countDegreeNotBacktracked(6);

// used in tick() and tock()
clock_t c_start = clock();
uint64_t cycles_start = rdtsc();

// some auxiliary functions mostly for printing data structures or applying simple operations on vectors
// -------------------------------------------------------

struct comparator {
	bool operator()(tuple<tuple<byte, byte>, int> i, tuple<tuple<byte, byte>, int> j) {
		return get<1>(i) > get<1>(j);
	}
};

std::ostream& operator<<(std::ostream &strm, const vector<byte> &vec) {
	string s("");
	for (unsigned int var = 0; var < vec.size(); ++var) {
		s += std::to_string((int)vec[var]);
		if (var != vec.size()-1) {
			s += ",";
		}
	}
	return strm << "(" << s << ")";
}

std::ostream& operator<<(std::ostream &strm, const vector<bool> &vec) {
	string s("");
	for (unsigned int var = 0; var < vec.size(); ++var) {
		s += std::to_string((bool)vec[var]);
		if (var != vec.size() - 1) {
			s += ",";
		}
	}
	return strm << "(" << s << ")";
}

vector<bool> operator|(vector<bool>& v1, const vector<bool>& v2) {
	vector<bool> res(v1.size());
	for (unsigned int i = 0; i < res.size(); i++) {
		res[i] = v1[i] | v2[i];
	}
	return res;
}




std::ostream& operator<<(std::ostream &strm, const vector<vector<byte>> &vec) {
	string s("");
	for (unsigned int var = 0; var < vec.size(); ++var) {
		strm << vec[var];
	}
	return strm;
}

std::ostream& operator<<(std::ostream &strm, const byte &byte) {
  return strm << std::to_string((int)byte);
}

std::ostream& operator<<(std::ostream &strm, const tuple<byte,byte> &tuple) {
	return strm << "(" << std::to_string((int)get<0>(tuple)) << "," << std::to_string((int)get<1>(tuple))  << ")";
}



std::ostream& operator<<(std::ostream &strm, const SudokuGrid &grid) {
	string s("");
	for (unsigned int row = 0; row < pow(grid.n,2); ++row) {
		for (int col = 0; col < pow(grid.n,2); ++col) {
			s += std::to_string((int)grid.grid[row][col]);
			if (col != pow(grid.n,2)-1) {
				s += ",";
			}

		}
		s += "\n";
	}
  return strm << s << endl;
}

std::ostream& operator<<(std::ostream &strm, const vector<SudokuGrid> &grid) {
	for (unsigned int i = 0; i < grid.size(); ++i) {
		strm << grid[i];
	}
  return strm;
}
// -------------------------------------------------------




// represents an array of domains; one domain for every cell in a Sudoku grid
class SudokuGridDomain
{
	public:
		// builds domain with every element
		SudokuGridDomain(int n) {
			domain.resize((int)pow(n,2));
			for (int i = 0; i < pow(n,2); ++i) {
				domain[i].resize((int)pow(n,2));
				for (int j = 0; j < pow(n,2); ++j) {
					for (int k = 1; k <= pow(n,2); ++k) {
						domain[i][j].push_back(k);
					}
				}
			}
		}

		// cells in 0..80 format
		SudokuGridDomain(const SudokuGrid& grid, const vector<byte>& cells) {
			domain.resize((int)pow(grid.n,2));
			for (int row = 0; row < pow(grid.n,2); ++row) {
				domain[row].resize((int)pow(grid.n,2));
				for (int col = 0; col < pow(grid.n,2); ++col) {
					bool found = false;
					for (unsigned int k = 0; k < cells.size(); ++k) {
						if (row*pow(grid.n, 2) + col == cells[k]) {
							found = true;
							for (int j = 1; j <= pow(grid.n, 2); ++j) {
								domain[row][col].push_back(j);
							}
							break;
						}

					}
					if (!found) {
						domain[row][col].push_back(grid.grid[row][col]);
					}
				}
			}
		}

		SudokuGridDomain(const SudokuGrid& grid) {
			domain.resize((int)pow(grid.n,2));
			for (int row = 0; row < pow(grid.n,2); ++row) {
				domain[row].resize((int)pow(grid.n,2));
				for (int col = 0; col < pow(grid.n,2); ++col) {
					if (grid.grid[row][col] == 0) {
						for (int j = 1; j <= pow(grid.n, 2); ++j) {
							domain[row][col].push_back(j);
						}
					}
					else {
						domain[row][col].push_back(grid.grid[row][col]);
					}
				}
			}
		}

		std::vector<std::vector<vector<byte> > > domain;
		std::vector<vector<byte> > operator [](int i) const {return domain[i];}
		std::vector<vector<byte> > & operator [](int i) {return domain[i];}
};

// solves a given Sudoku grid with using a backtracking algorithm which uses forward checking and an ascending variable ordering
vector<SudokuGrid> backTrackingWithForwardChecking(vector<tuple<byte, byte>>& V, int n, const SudokuGrid& grid, vector<SudokuGrid>& array, SudokuGridDomain& domain, 
	const bool findMultipleSolutions, const bool cancelWhenMultipleFound) {
	unsigned int minDomainSize = 10000;
	int row, col, index = 0, rowSave, colSave;
	for (unsigned int i = 0; i < V.size(); ++i) {
		if (domain[get<0>(V[i])][get<1>(V[i])].size() < minDomainSize) {
			minDomainSize = domain[get<0>(V[i])][get<1>(V[i])].size();
			row = get<0>(V[i]);
			rowSave = row;
			col = get<1>(V[i]);
			colSave = col;
			index = i;
		}
	}
	V.erase(V.begin() + index);
//	cout << col << " " << row << endl;
	const vector<byte>& branches = domain[row][col];
	vector<SudokuGrid> grids;
	grids.resize(branches.size(), grid);
	vector<SudokuGridDomain> domains;
	domains.resize(branches.size(), domain);
	vector<vector<tuple<byte, byte>>> Vs;
	Vs.resize(branches.size(), V);
	for (unsigned int i = 0; i < branches.size(); ++i) {
		vector<tuple<byte, byte>> VOneElement;
		byte chosenByte = branches[i];
jumppoint:grids[i].grid[row][col] = chosenByte;
		if (Vs[i].empty()) {
			array.push_back(grids[i]);
			if (cancelWhenMultipleFound && !((findMultipleSolutions && array.size() < 2) || array.size() == 0)) {
				return array;
			}
			else {
				row = rowSave;
				col = colSave;
				continue;
			}
		}
		// forward checking
		bool checkOK = true;
		// update column
		for (int coli = 0; coli < pow(n,2); ++coli) {
			vector<byte>& chosenDomain = domains[i][row][coli];
			chosenDomain.erase(std::remove(chosenDomain.begin(), chosenDomain.end(), chosenByte), chosenDomain.end());
			if (chosenDomain.size() == 0 && grids[i].grid[row][coli] == 0) {
				checkOK = false;
				break;
			}
			else if (chosenDomain.size() == 1 && grids[i].grid[row][coli] == 0) {
				VOneElement.push_back(tuple<byte, byte>(row, coli));
			}
		}

		if (checkOK)
		{
			// update row
			for (int rowi = 0; rowi < pow(n, 2); ++rowi) {
				vector<byte>& chosenDomain = domains[i][rowi][col];
				chosenDomain.erase(std::remove(chosenDomain.begin(), chosenDomain.end(), chosenByte), chosenDomain.end());
				if (chosenDomain.size() == 0 && grids[i].grid[rowi][col] == 0) {
					checkOK = false;
					break;
				}
				else if (chosenDomain.size() == 1 && grids[i].grid[rowi][col] == 0) {
					VOneElement.push_back(tuple<byte, byte>(rowi, col));
				}
			}
		}
		if (checkOK)
		{
			// update square
			for (int rowOffset = 0; rowOffset < n; ++rowOffset) {
				for (int colOffset = 0; colOffset < n; ++colOffset) {
					//				cout << (int)((row/n)*n+rowOffset) << " " << (int)((col/n)*n+colOffset) << endl;
					vector<byte>& chosenDomain = domains[i][(int)((row / n)*n + rowOffset)][(int)((col / n)*n + colOffset)];
					chosenDomain.erase(std::remove(chosenDomain.begin(), chosenDomain.end(), chosenByte), chosenDomain.end());
					if (chosenDomain.size() == 0 && grids[i].grid[(int)((row / n)*n + rowOffset)][(int)((col / n)*n + colOffset)] == 0) {
						checkOK = false;
						break;
					}
					else if (chosenDomain.size() == 1 && grids[i].grid[(int)((row / n)*n + rowOffset)][(int)((col / n)*n + colOffset)] == 0) {
						VOneElement.push_back(tuple<byte, byte>((int)((row / n)*n + rowOffset), (int)((col / n)*n + colOffset)));
					}
				}
			}

		}

		// assign all variables which only have 1 element in their domains
		if (checkOK && !VOneElement.empty()) {
			tuple<byte, byte> t = VOneElement.back();
			Vs[i].erase(std::remove(Vs[i].begin(), Vs[i].end(), t), Vs[i].end());
			VOneElement.pop_back();
			row = get<0>(t);
			col = get<1>(t);
			chosenByte = domains[i][row][col][0];
			goto jumppoint;
		}
		row = rowSave;
		col = colSave;

		if (checkOK) {
			if (!cancelWhenMultipleFound || (findMultipleSolutions && array.size() < 2) || array.size() == 0) {
				backTrackingWithForwardChecking(Vs[i], n, grids[i], array, domains[i], findMultipleSolutions, cancelWhenMultipleFound);
			}
			if (cancelWhenMultipleFound && !((findMultipleSolutions && array.size() < 2) || array.size() == 0)) {
				return array;
			}

		}
	}
	return array;
}

// runs all blueprints over a already transformed grid and returns the matched blueprints
void checkGridForPatterns(const SudokuGrid& grid, const UnavPat * blueprintArray, int arraycount, vector<vector<byte> >& blueprints) {
	for (int i = 0; i < arraycount; ++i) {
		UnavPat blueprint = blueprintArray[i];
		if (grid.n == 2) {
			if (!blueprint.shrink(grid.n)) {
				continue;
			}
		}
		bool patternMatches = true;
		for (int digit = 0; digit < blueprint.digits; ++digit) {
			SmallSet set = blueprint.groups[digit];
			byte referenceDigit = grid.grid[set.data[0] / 10 - 1]
									   [set.data[0] % 10 - 1];
			for (int var = 1; var < set.size; ++var) {
				if (grid.grid[set.data[var] / 10 - 1]
						 [set.data[var] % 10 - 1]
						!= referenceDigit) {
					patternMatches = false;
					goto skip;
				}
			}
		}
		skip:if (patternMatches) {
			blueprints.push_back(blueprint.toByteVector());
		}
	}
}

// finds unev sets which are in two or more bands
void findUnavoidableSetsInTwoOrMoreBands(const SudokuGrid& grid, vector<vector<byte> >& correctlyTransformedBlueprints) {
	uint64_t count2x2 = 0;
	uint64_t count2x3 = 0;
	uint64_t count3x3 = 0;
	for (int transposed = 0; transposed <= 1; ++transposed) {
		SudokuGrid grid1 = grid;
		if (transposed) {
			grid1.transpose();
		}
		for (unsigned int tbands = 0; tbands < factorial(grid.n); ++tbands) {
			SudokuGrid grid2 = grid1;
			grid2.permuteBands(tbands,false);
			for (unsigned int trowstop = 0; trowstop < factorial(grid.n); ++trowstop) {
				SudokuGrid grid3 = grid2;
				grid3.permuteRows(0,trowstop,false);
				for (unsigned int tstacks = 0; tstacks < factorial(grid.n); ++tstacks) {
					SudokuGrid grid4 = grid3;
					grid4.permuteStacks(tstacks,false);
					for (unsigned int tleftcols = 0; tleftcols < factorial(grid.n); ++tleftcols) {
						SudokuGrid grid5 = grid4;
						grid5.permuteCols(0,tleftcols,false);
						for (int tmiddlecols = 0; tmiddlecols < grid.n; ++tmiddlecols) {
							SudokuGrid grid6 = grid5;
							grid6.swapCols(grid.n,grid.n+tmiddlecols);
							if (grid6.grid[0][0] == grid6.grid[1][grid.n]) {
								for (unsigned int trowsmid = 0; trowsmid < factorial(grid.n); trowsmid+=(grid.n-1)) {
									SudokuGrid grid7 = grid6;
									grid7.permuteRows(1,trowsmid,false);
									if (grid7.grid[1][0] == grid7.grid[grid.n][1] || (grid.n > 2 && grid7.grid[1][2] == grid7.grid[3][1]) ||
										(grid.n > 2 && grid7.grid[1][2] == grid7.grid[grid.n][0])) {
										for (unsigned int trowsbot = 0; (grid.n > 2 || trowsbot == 0) && trowsbot < factorial(grid.n); trowsbot+=(grid.n-1)) {
											SudokuGrid grid8 = grid7;
											if (grid.n > 2) {
												grid8.permuteRows(2, trowsbot, false);
											}
											for (int trightcols = 0; (grid.n > 2 || trightcols == 0) && (trightcols < grid.n); ++trightcols) {
												SudokuGrid transformedGrid = grid8;
												if (grid.n > 2) {
													transformedGrid.swapCols(grid.n*2,grid.n*2+trightcols);
												}
												vector<vector<byte> > notYetTransformedPatterns = vector<vector<byte> >();

												// the following if cascade is copied from the programme checker
												// it can be found on http://www.math.ie/checker.html
												// the types are referring to figure 9 of their paper:
												// "There Is No 16-Clue Sudoku:..."
												// whole title and reference can be found in the first lines of this file
												int type = 0;           // type I
												if (grid.n > 2 && transformedGrid.grid[1][2] == transformedGrid.grid[3][1])
													type = 3;            // type II
												else if ( transformedGrid.grid[1][0] == transformedGrid.grid[grid.n][1])
												{
													type = 1;            // type III
													if (grid.n > 2 && transformedGrid.grid[1][2] == transformedGrid.grid[grid.n][0])
														type = 2;         // type I + type III
												}
												if (trowsbot == 0 && trightcols == 0) {
													int temp = notYetTransformedPatterns.size();
													checkGridForPatterns(transformedGrid, unavpat2x2[type], cntpat2x2[type], notYetTransformedPatterns);
													count2x2 += notYetTransformedPatterns.size() - temp;
												}
												if (grid.n > 2) {
													if (trowsbot == 0) {
														int temp = notYetTransformedPatterns.size();
														checkGridForPatterns(transformedGrid, unavpat2x3[type], cntpat2x3[type], notYetTransformedPatterns);
														count2x3 += notYetTransformedPatterns.size() - temp;
													}
													int temp = notYetTransformedPatterns.size();
													checkGridForPatterns(transformedGrid, unavpat3x3[type], cntpat3x3[type], notYetTransformedPatterns);
													count3x3 += notYetTransformedPatterns.size() - temp;
												}

												for (unsigned int i = 0; i < notYetTransformedPatterns.size(); ++i) {
													SudokuGrid patterngrid = SudokuGrid(grid.n, notYetTransformedPatterns[i]);
												//								cout << patterngrid << endl;
													if (grid.n > 2) {
														patterngrid.swapCols(grid.n*2,grid.n*2+trightcols);
														patterngrid.permuteRows(2, trowsbot, true);
													}
													patterngrid.permuteRows(1,trowsmid,true);
													patterngrid.swapCols(grid.n,grid.n+tmiddlecols);
													patterngrid.permuteCols(0,tleftcols,true);
													patterngrid.permuteStacks(tstacks,true);
													patterngrid.permuteRows(0,trowstop,true);
													patterngrid.permuteBands(tbands,true);
													if (transposed) {
														patterngrid.transpose();
													}
													correctlyTransformedBlueprints.push_back(patterngrid.getNotZeroCells());
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	cout << "found " << count2x2 << " 2x2 patterns" << endl;
	cout << "found " << count2x3 << " 2x3 patterns" << endl;
	cout << "found " << count3x3 << " 3x3 patterns" << endl;
}

// finds unev sets which are occupying only one Band
void findUnavoidableSetsInOneBand(const SudokuGrid& grid, vector<vector<byte> >& correctlyTransformedBlueprints) {
	uint64_t count1x2 = 0;
	uint64_t count1x3 = 0;
	for (int transposed = 0; transposed <= 1; ++transposed) {
		SudokuGrid grid1 = grid;
		if (transposed) {
			grid1.transpose();
		}
		for (int tbands = 0; tbands < grid.n; ++tbands) {
			SudokuGrid grid2 = grid1;
			grid2.swapBands(0,tbands);
			for (unsigned int trowstop = 0; trowstop < factorial(grid.n); ++trowstop) {
				SudokuGrid grid3 = grid2;
				grid3.permuteRows(0,trowstop,false);
				for (unsigned int tstacks = 0; tstacks < factorial(grid.n); ++tstacks) {
					SudokuGrid grid4 = grid3;
					grid4.permuteStacks(tstacks,false);
					for (unsigned int tleftcols = 0; tleftcols < factorial(grid.n); ++tleftcols) {
						SudokuGrid grid5 = grid4;
						grid5.permuteCols(0,tleftcols,false);
						for (int tmiddlecols = 0; tmiddlecols < grid.n; ++tmiddlecols) {
							SudokuGrid grid6 = grid5;
							grid6.swapCols(grid.n,grid.n+tmiddlecols);
							if (grid6.grid[0][0] == grid6.grid[1][grid.n]) {
								vector<vector<byte> > notYetTransformedPatterns = vector<vector<byte> >();
								checkGridForPatterns(grid6, unavpat1x2, cntpat1x2, notYetTransformedPatterns);
								count1x2 += notYetTransformedPatterns.size();

								for (unsigned int i = 0; i < notYetTransformedPatterns.size(); ++i) {
									SudokuGrid patterngrid = SudokuGrid(grid.n, notYetTransformedPatterns[i]);
									patterngrid.swapCols(grid.n,grid.n+tmiddlecols);
									patterngrid.permuteCols(0,tleftcols,true);
									patterngrid.permuteStacks(tstacks,true);
									patterngrid.permuteRows(0,trowstop,true);
									patterngrid.swapBands(0,tbands);
									if (transposed) {
										patterngrid.transpose();
									}
									correctlyTransformedBlueprints.push_back(patterngrid.getNotZeroCells());
								}
								if (grid.n > 2) {
									for (int trightcols = 0; trightcols < grid.n; ++trightcols) {
										SudokuGrid grid7 = grid6;
										grid7.swapCols(grid.n*2,grid.n*2+trightcols);
										vector<vector<byte> > notYetTransformedPatterns = vector<vector<byte> >();
										checkGridForPatterns(grid7, unavpat1x3, cntpat1x3, notYetTransformedPatterns);
										count1x3 += notYetTransformedPatterns.size();

										for (unsigned int i = 0; i < notYetTransformedPatterns.size(); ++i) {
											SudokuGrid patterngrid = SudokuGrid(grid.n, notYetTransformedPatterns[i]);
											patterngrid.swapCols(grid.n*2,grid.n*2+trightcols);
											patterngrid.swapCols(grid.n,grid.n+tmiddlecols);
											patterngrid.permuteCols(0,tleftcols,true);
											patterngrid.permuteStacks(tstacks,true);
											patterngrid.permuteRows(0,trowstop,true);
											patterngrid.swapBands(0,tbands);
											if (transposed) {
												patterngrid.transpose();
											}
											correctlyTransformedBlueprints.push_back(patterngrid.getNotZeroCells());
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	cout << "found " << count1x2 << " 1x2 patterns" << endl;
	cout << "found " << count1x3 << " 1x3 patterns" << endl;
}

// returns all minimal unavoidable sets of size <= 12; cells are encoded from 0 to 80
vector<vector<byte>> findUnavoidableSets(SudokuGrid grid) {
	vector<vector<byte> > correctlyTransformedBlueprints;
	findUnavoidableSetsInOneBand(grid, correctlyTransformedBlueprints);
	findUnavoidableSetsInTwoOrMoreBands(grid, correctlyTransformedBlueprints);
	// eliminate duplicates
	int duplicateCount = 0;
	for (unsigned int i = 0; i < correctlyTransformedBlueprints.size(); ++i) {
		for (unsigned int j = 0; j < correctlyTransformedBlueprints.size(); ++j) {
			if (i == j) {
				continue;
			}
			if (correctlyTransformedBlueprints[i].size() == correctlyTransformedBlueprints[j].size()) {
				bool identical = true;
				for (unsigned int k = 0; k < correctlyTransformedBlueprints[i].size(); ++k) {
					if (correctlyTransformedBlueprints[i][k] != correctlyTransformedBlueprints[j][k]) {
						identical = false;
						break;
					}
				}
				if (identical) {
					correctlyTransformedBlueprints.erase(correctlyTransformedBlueprints.begin() + j);
					duplicateCount++;
					j--;
				}
			}
		}
	}
	cout << "there were " << duplicateCount << " duplicates eliminated" << endl;
	for (unsigned int i = 0; i < correctlyTransformedBlueprints.size(); i++) {
		for (unsigned int j = 0; j < correctlyTransformedBlueprints[i].size(); j++) {
			correctlyTransformedBlueprints[i][j] = (int)((correctlyTransformedBlueprints[i][j] / 10 - 1) * pow(grid.n, 2) + ((correctlyTransformedBlueprints[i][j] % 10) - 1));
		}
	}
	return correctlyTransformedBlueprints;
}

// checks if there is a solution and if requested returns every Sudoku solution grid which is completing the given grid
vector<SudokuGrid> solveSudoku(const SudokuGrid& grid, SudokuGridDomain& domain, bool findMultipleSolutions, bool cancelWhenMultipleFound) {
	vector<tuple<byte, byte>> V;
	for (int row = 0; row < pow(grid.n,2); ++row) {
		for (int col = 0; col < pow(grid.n,2); ++col) {
			V.push_back(tuple<byte, byte>(row, col));
		}
	}
	vector<SudokuGrid> gridVector;
	return backTrackingWithForwardChecking(V, grid.n, grid, gridVector, domain, findMultipleSolutions, cancelWhenMultipleFound);
}



// debugging function: checks whether a potential unavoidable set is actually an unavoidable set
bool isUnavoidableSetValid(vector<byte> set, SudokuGrid grid) {
	SudokuGridDomain domain(grid, set);
	vector<SudokuGrid> solutionGrids = solveSudoku(grid, domain, true, true);
	if (solutionGrids.size() > 1) {
		return true;
	}
	return false;
}

// creates all Sudoku solution grids for Sudokus of size nxn
vector<SudokuGrid> createSolutionGrids(int n) {
	SudokuGrid grid(n);
	SudokuGridDomain domain(n);
	return solveSudoku(grid, domain, true, false);
}



// -1 for all
vector<SudokuGrid> read17ClueSudokus(int amount) {
	vector<SudokuGrid> vec;
	ifstream input("sudoku17-ml.txt");
	string line;
	int index = 0;
	for( string line; getline( input, line ); ) {
		cout << line << endl;
		vector<byte> digits;
		for(char& c : line) {
		    digits.push_back((c-48));
		}
	    vec.push_back(SudokuGrid(digits));
	    index++;
	    if (amount != -1 && index >= amount) {
			break;
		}
	}
	return vec;
}

// returns 0 if offset is out of file range
SudokuGrid readSpecific17ClueSudoku(int offset) {
	ifstream input("sudoku17-ml.txt");
	if (input.fail()) {
		cout << "failed to load file" << endl;
	}
	string line;
	int index = 0;
	for (string line; getline(input, line); ) {
		if (index != offset) {
			index++;
			continue;
		}
		else {
			vector<byte> digits;
			for (char& c : line) {
				digits.push_back((c - 48));
			}
			return SudokuGrid(digits);
		}
	}
	return 0;
}

// for every cells which unav sets are hit
vector<boost::dynamic_bitset<>> initHittingVectors(SudokuGrid& grid, vector<vector<byte> >& unavoidableSets) {
	// initialize traditional hittingvec
	vector<boost::dynamic_bitset<>> hittingvector((int)pow(grid.n, 4));
	for (int var = 0; var < pow(grid.n, 4); ++var) {
		boost::dynamic_bitset<> db{ unavoidableSets.size(), 0 };
		hittingvector[var] = db;
	}
	for (int i = 0; i < pow(grid.n, 4); i++) {
		for (unsigned int j = 0; j < unavoidableSets.size(); j++) {
			for (unsigned int k = 0; k < unavoidableSets[j].size(); k++) {
				if (i == unavoidableSets[j][k]) {
					hittingvector[i][j] = 1;
				}
			}
		}
	}
	return hittingvector;
}

// debugging function: checks whether a potential hitting set actually is a valid hitting set
bool doesHitsetHitAllUnavoidableSets(vector<byte>& hitset, vector<vector<byte> >& unavoidableSets) {
	for (unsigned int i = 0; i < unavoidableSets.size(); i++) {
		for (unsigned int j = 0; j < unavoidableSets[i].size(); j++) {
			for (unsigned int k = 1; k < hitset.size(); k++) {
				if (unavoidableSets[i][j] == hitset[k]) {
					goto jump;
				}
			}
		}
		return false;
	jump:1 + 1;
	}
	return true;
}

// creates every possible Sudoku which can be constructed out of the hitting set and checks whether there is any which is actually valid
bool generatePuzzles(int n, int k, const vector<byte>& hitset, const boost::dynamic_bitset<>& deadvec, const SudokuGrid& grid) {
	if (k == n) {
		countHittingSets++;
		// cout << n << "-clue-hitset found now checking for multiple solutions" << endl << hitset << endl;
		// check here whether hitset has only one sudoku grid; basic hit set algorithm is implemented though
		clock_t temp = clock();
		SudokuGrid checkGrid(grid, hitset);
		SudokuGridDomain checkDomain(checkGrid);
		vector<SudokuGrid> solutionGrids = solveSudoku(checkGrid, checkDomain, true, true);
		timeSpentCheckingCandidates += clock() - temp;
		// cout << "check done" << endl;
		if (solutionGrids.size() == 1) {
			return true;
		}
		else {
			return false;
		}
	}

	for (unsigned int i = 0; i < deadvec.size(); i++) {
		if (deadvec[i] == 0) {
			vector<byte> hitset2(hitset);
			boost::dynamic_bitset<> deadvec2(deadvec);
			hitset2[k + 1] = i;
			deadvec2[i] = 1;
			if (generatePuzzles(n, k + 1, hitset2, deadvec2, grid)) {
				return true;
			}
		}
	}
	return false;
}

// after drawing 5 clues only 1536 unhit higher degree unavoidable sets are kept
vector<vector<boost::dynamic_bitset<>>> consolidateHigherHittingVectors(vector<boost::dynamic_bitset<>>& higherStateVec, const SudokuGrid& grid, int k,
	const vector<vector<boost::dynamic_bitset<>>>& higherHittingsVecs, vector<vector<boost::dynamic_bitset<>>>& consolidatedHigherHittingVecs) {
	for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
		consolidatedHigherHittingVecs[5][clue].reset();
		consolidatedHigherHittingVecs[4][clue].reset();
		consolidatedHigherHittingVecs[3][clue].reset();
		consolidatedHigherHittingVecs[2][clue].reset();
	}
	int j = 0;
	// degree 5
	for (unsigned int i = 0; i < higherHittingsVecs[5][0].size(); i++) {
		if (higherStateVec[5][i] == false) { // if not hit yet add it to new unav set set
			for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
				consolidatedHigherHittingVecs[5][clue][j] = higherHittingsVecs[5][clue][i];
			}
			j++;
			if (j == 1536) {
				break;
			}
		}
	}
	if (j != 1536) {
		for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
			for (int restindex = j; restindex < 1536; restindex++) {
				consolidatedHigherHittingVecs[5][clue][restindex] = true;
			}
		}
	}
	// degree 4
	j = 0;
	for (unsigned int i = 0; i < higherHittingsVecs[4][0].size(); i++) {
		if (higherStateVec[4][i] == false) { // if not hit yet add it to new unav set set
			//debugAssignVec[j] = i;
			for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
				consolidatedHigherHittingVecs[4][clue][j] = higherHittingsVecs[4][clue][i];
			}
			j++;
			if (j == 1536) {
				break;
			}
		}
	}
	if (j != 1536) {
		for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
			for (int restindex = j; restindex < 1536; restindex++) {
				consolidatedHigherHittingVecs[4][clue][restindex] = true;
			}
		}
	}
	// degree 3
	j = 0;
	for (unsigned int i = 0; i < higherHittingsVecs[3][0].size(); i++) {
		if (higherStateVec[3][i] == false) { // if not hit yet add it to new unav set set
			for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
				consolidatedHigherHittingVecs[3][clue][j] = higherHittingsVecs[3][clue][i];
			}
			j++;
			if (j == 1536) {
				break;
			}
		}
	}
	if (j != 1536) {
		for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
			for (int restindex = j; restindex < 1536; restindex++) {
				consolidatedHigherHittingVecs[3][clue][restindex] = true;
			}
		}
	}
	// degree 2
	j = 0;
	for (unsigned int i = 0; i < higherHittingsVecs[2][0].size(); i++) {
		if (higherStateVec[2][i] == false) { // if not hit yet add it to new unav set set
			for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
				consolidatedHigherHittingVecs[2][clue][j] = higherHittingsVecs[2][clue][i];
			}
			j++;
			if (j == 1536) {
				break;
			}
		}
	}
	if (j != 1536) {
		for (int clue = 0; clue < grid.n*grid.n*grid.n*grid.n; clue++) {
			for (int restindex = j; restindex < 1536; restindex++) {
				consolidatedHigherHittingVecs[2][clue][restindex] = true;
			}
		}
	}
	return consolidatedHigherHittingVecs;
}

// tries to construct a hitting set recursively if one is found it checks whether it also corresponds to a valid Sudoku
bool addClue(int cluesDrawn, int n, int minimumIndex, vector<vector<boost::dynamic_bitset<> > >& deadclues, vector<boost::dynamic_bitset<> >& deadvec, 
	vector<boost::dynamic_bitset<> >& hittingvec, vector<boost::dynamic_bitset<> >& statevec, vector<byte>& hitset, vector<vector<byte> >& unavoidableSets, 
	const SudokuGrid& grid, const vector<vector<boost::dynamic_bitset<>>>& higherHittingsVecs, vector<vector<boost::dynamic_bitset<>>>& higherStateVec, 
	const vector<vector<vector<byte>>>& higherDimUnav, vector<vector<boost::dynamic_bitset<>>>& consolidatedHigherHittingVecs) {

	// check for backtracking at the start
	if (cluesDrawn == 0) {
		if ((maxClues >= 1 && maxClues <= 4) && higherDimUnav[maxClues + 1].size() > 0) {
			countDegreeBacktracked[maxClues + 1]++;
			return false;
		}
		if (maxClues >= 1 && maxClues <= 4) {
			countDegreeNotBacktracked[maxClues + 1]++;
		}
	}

	bool allHit = true; // if all unavoidable sets are hit

	// search for next unhit ua set
	for (unsigned int i = minimumIndex; i < statevec[cluesDrawn].size(); i++) {
		if (statevec[cluesDrawn][i] == 0) {
			minimumIndex = i;
			allHit = false;
			break;
		}
	}
	
	if (allHit) {
		if (cluesDrawn != n) {
			countSmallerHittingSets++;
		}
		// cout << "works: " << doesHitsetHitAllUnavoidableSets(hitset, unavoidableSets); seems to work so uncommented for final run
		bool found16ClueSudoku = generatePuzzles(n, cluesDrawn, hitset, deadvec[cluesDrawn], grid);
		if (found16ClueSudoku) {
			cout << "found a valid " << maxClues << " clue Sudoku (this should not happen)" << endl;
			cout << grid << endl;
			cout << hitset << endl;
		}
		return found16ClueSudoku;
	}
	// already maxClues clues but haven't hit all unavoidable sets yet
	else if (cluesDrawn == n) {
		countLastLayerNotAllHit++;
		return false;
	}
	vector<byte>& U = unavoidableSets[minimumIndex];
	for (unsigned int clueIndex = 0; clueIndex < U.size(); clueIndex++) {
		byte c = U[clueIndex];
		// is clue still alive
		if (deadvec[cluesDrawn][c] == 0) {
			countNotIgnored[cluesDrawn]++;
			hitset[cluesDrawn + 1] = c;

			// update which sets are already hit
			statevec[cluesDrawn + 1] = statevec[cluesDrawn];
			statevec[cluesDrawn + 1] |= hittingvec[c];
			// update current deadvec
			deadvec[cluesDrawn + 1] = deadvec[cluesDrawn];
			deadvec[cluesDrawn + 1] |= deadclues[minimumIndex][clueIndex];

			// update higher state vec if not consolidated (because if we did we want to keep the zero vector, because none of the new ua sets have been hit)
			if (cluesDrawn + 1 != 5) {
				// for degree 5
				if (cluesDrawn + 1 <= n - 4) {
					higherStateVec[5][cluesDrawn + 1] = higherStateVec[5][cluesDrawn];
					if (cluesDrawn + 1 <= 5) {
						higherStateVec[5][cluesDrawn + 1] |= higherHittingsVecs[5][c];
					}

					else {
						higherStateVec[5][cluesDrawn + 1] |= consolidatedHigherHittingVecs[5][c];
					}
				}
				// for degree 4
				if (cluesDrawn + 1 <= n - 3) {
					higherStateVec[4][cluesDrawn + 1] = higherStateVec[4][cluesDrawn];
					if (cluesDrawn + 1 <= 5) {
						higherStateVec[4][cluesDrawn + 1] |= higherHittingsVecs[4][c];
					}

					else {
						higherStateVec[4][cluesDrawn + 1] |= consolidatedHigherHittingVecs[4][c];
					}
				}
				// for degree 3
				if (cluesDrawn + 1 <= n - 2) {
					higherStateVec[3][cluesDrawn + 1] = higherStateVec[3][cluesDrawn];
					if (cluesDrawn + 1 <= 5) {
						higherStateVec[3][cluesDrawn + 1] |= higherHittingsVecs[3][c];
					}

					else {
						higherStateVec[3][cluesDrawn + 1] |= consolidatedHigherHittingVecs[3][c];
					}
				}
				// for degree 2
				if (cluesDrawn + 1 <= n - 1) {
					higherStateVec[2][cluesDrawn + 1] = higherStateVec[2][cluesDrawn];
					if (cluesDrawn + 1 <= 5) {
						higherStateVec[2][cluesDrawn + 1] |= higherHittingsVecs[2][c];
					}

					else {
						higherStateVec[2][cluesDrawn + 1] |= consolidatedHigherHittingVecs[2][c];
					}
				}
			}
			else {
				// after drawing 5 clues make higherHittingVec smaller, because already most of them have been hit
				vector<boost::dynamic_bitset<>> temp(6);
				temp[5] = higherStateVec[5][cluesDrawn];
				temp[5] |= higherHittingsVecs[5][c];
				temp[4] = higherStateVec[4][cluesDrawn];
				temp[4] |= higherHittingsVecs[4][c];
				temp[3] = higherStateVec[3][cluesDrawn];
				temp[3] |= higherHittingsVecs[3][c];
				temp[2] = higherStateVec[2][cluesDrawn];
				temp[2] |= higherHittingsVecs[2][c];
				consolidatedHigherHittingVecs = consolidateHigherHittingVectors(temp, grid, cluesDrawn + 1, higherHittingsVecs, consolidatedHigherHittingVecs);
			}

			// if backtracking available
			if ((cluesDrawn + 1 == n - 4 && !higherStateVec[5][cluesDrawn + 1].all()) || (cluesDrawn + 1 == n - 3 && !higherStateVec[4][cluesDrawn + 1].all()) || (cluesDrawn + 1 == n - 2 && !higherStateVec[3][cluesDrawn + 1].all()) || (cluesDrawn + 1 == n - 1 && !higherStateVec[2][cluesDrawn + 1].all())) {
				countDegreeBacktracked[maxClues - (cluesDrawn)]++;
				continue;
			}
			else if ((maxClues - (cluesDrawn + 1)) >= 1 && (maxClues - (cluesDrawn + 1) <= 4)) {
				countDegreeNotBacktracked[maxClues - (cluesDrawn)]++;
			}

			
			// add remaining clues, maybe remove if to save some computing power
			if (addClue(cluesDrawn + 1, n, minimumIndex + 1, deadclues, deadvec, hittingvec, statevec, hitset, unavoidableSets, grid, higherHittingsVecs,
				higherStateVec, higherDimUnav, consolidatedHigherHittingVecs)) {
				return true;
			}
		}
		else {
			countIgnored[cluesDrawn]++;
		}
	}

	return false;
}

// creates higher degree unavoidable sets from degree 2 to degree 5 - eats up a lot of initial computation
vector<vector<vector<byte>>> createHigherDimensionalUnavoidableSets(const vector<vector<byte> >& unavoidableSets, int n) {
	vector<vector<vector<byte>>> higherDimUnav;
	higherDimUnav.resize(6);
	for (int i = n; i >= 0 && i < unavoidableSets.size(); i++) {
		for (unsigned int j = n-1; j <= i-1; j++) {
			for (unsigned int j1 = 0; j1 < unavoidableSets[j].size(); j1++) {
				for (unsigned int i1 = 0; i1 < unavoidableSets[i].size(); i1++) {
					if (unavoidableSets[j][j1] == unavoidableSets[i][i1]) {
						goto next;
					}
				}
			}
			// found 2 disjunctive sets
			{
				vector<byte> uaset;
				uaset.reserve(unavoidableSets[i].size() + unavoidableSets[j].size());
				uaset.insert(uaset.end(), unavoidableSets[i].begin(), unavoidableSets[i].end());
				uaset.insert(uaset.end(), unavoidableSets[j].begin(), unavoidableSets[j].end());
				higherDimUnav[2].push_back(uaset);
			}
			for (int k = n-2; k >= 0 && k <= j-1; k++) {
				for (unsigned int k1 = 0; k1 < unavoidableSets[k].size(); k1++) {
					for (unsigned int i1 = 0; i1 < unavoidableSets[i].size(); i1++) {
						if (unavoidableSets[k][k1] == unavoidableSets[i][i1]) {
							goto next2;
						}
					}
					for (unsigned int j1 = 0; j1 < unavoidableSets[j].size(); j1++) {
						if (unavoidableSets[k][k1] == unavoidableSets[j][j1]) {
							goto next2;
						}
					}
				}
				// found 3 disjunctive sets
				{
					vector<byte> uaset;
					uaset.reserve(unavoidableSets[i].size() + unavoidableSets[j].size() + unavoidableSets[k].size());
					uaset.insert(uaset.end(), unavoidableSets[i].begin(), unavoidableSets[i].end());
					uaset.insert(uaset.end(), unavoidableSets[j].begin(), unavoidableSets[j].end());
					uaset.insert(uaset.end(), unavoidableSets[k].begin(), unavoidableSets[k].end());
					higherDimUnav[3].push_back(uaset);
				}
				for (int l = n-3; l >= 0 && l <= k-1; l++) {
					for (unsigned int l1 = 0; l1 < unavoidableSets[l].size(); l1++) {
						for (unsigned int i1 = 0; i1 < unavoidableSets[i].size(); i1++) {
							if (unavoidableSets[l][l1] == unavoidableSets[i][i1]) {
								goto next3;
							}
						}
						for (unsigned int j1 = 0; j1 < unavoidableSets[j].size(); j1++) {
							if (unavoidableSets[l][l1] == unavoidableSets[j][j1]) {
								goto next3;
							}
						}
						for (unsigned int k1 = 0; k1 < unavoidableSets[k].size(); k1++) {
							if (unavoidableSets[l][l1] == unavoidableSets[k][k1]) {
								goto next3;
							}
						}
					}
					// found 4 discunctive sets
					{
						vector<byte> uaset;
						uaset.reserve(unavoidableSets[i].size() + unavoidableSets[j].size() + unavoidableSets[k].size()
							+ unavoidableSets[l].size());
						uaset.insert(uaset.end(), unavoidableSets[i].begin(), unavoidableSets[i].end());
						uaset.insert(uaset.end(), unavoidableSets[j].begin(), unavoidableSets[j].end());
						uaset.insert(uaset.end(), unavoidableSets[k].begin(), unavoidableSets[k].end());
						uaset.insert(uaset.end(), unavoidableSets[l].begin(), unavoidableSets[l].end());
						higherDimUnav[4].push_back(uaset);
					}
					for (int m = n-4; m >= 0 && m <= l-1; m++) {
						for (unsigned int m1 = 0; m1 < unavoidableSets[m].size(); m1++) {
							for (unsigned int i1 = 0; i1 < unavoidableSets[i].size(); i1++) {
								if (unavoidableSets[m][m1] == unavoidableSets[i][i1]) {
									goto next4;
								}
							}
							for (unsigned int j1 = 0; j1 < unavoidableSets[j].size(); j1++) {
								if (unavoidableSets[m][m1] == unavoidableSets[j][j1]) {
									goto next4;
								}
							}
							for (unsigned int k1 = 0; k1 < unavoidableSets[k].size(); k1++) {
								if (unavoidableSets[m][m1] == unavoidableSets[k][k1]) {
									goto next4;
								}
							}
							for (unsigned int l1 = 0; l1 < unavoidableSets[i].size(); l1++) {
								if (unavoidableSets[m][m1] == unavoidableSets[l][l1]) {
									goto next4;
								}
							}
						}
						// found 5 disjunctive sets
						{
							vector<byte> uaset;
							uaset.reserve(unavoidableSets[i].size() + unavoidableSets[j].size() + unavoidableSets[k].size()
								+ unavoidableSets[l].size() + unavoidableSets[m].size());
							uaset.insert(uaset.end(), unavoidableSets[i].begin(), unavoidableSets[i].end());
							uaset.insert(uaset.end(), unavoidableSets[j].begin(), unavoidableSets[j].end());
							uaset.insert(uaset.end(), unavoidableSets[k].begin(), unavoidableSets[k].end());
							uaset.insert(uaset.end(), unavoidableSets[l].begin(), unavoidableSets[l].end());
							uaset.insert(uaset.end(), unavoidableSets[m].begin(), unavoidableSets[m].end());
							higherDimUnav[5].push_back(uaset);
						}
						next4:1 + 1;
					}
					next3:1 + 1;
				}
				next2:1 + 1;
			}
			next:1 + 1;
		}
	}
	return higherDimUnav;
}

// creates hitting vectors for the higher dimensional unavoidable sets from degree 2 to 5
// maps for each degree, each cell to a binary vector which is 1 at the i-th entry if this cell is contained in the i-th unavoidable set
vector<vector<boost::dynamic_bitset<>>> createHigherHittingsVecs(const vector<vector<vector<byte>>>& higherDimUnav, int cellCount) {
	vector<vector<boost::dynamic_bitset<>>> higherHittingsVecs;
	higherHittingsVecs.resize(6);
	// for degree 5
	higherHittingsVecs[5].resize(cellCount, boost::dynamic_bitset<> {higherDimUnav[5].size(), 0});
	for (unsigned int i = 0; i < higherDimUnav[5].size(); i++) {
		for (unsigned int j = 0; j < higherDimUnav[5][i].size(); j++) {
			higherHittingsVecs[5][higherDimUnav[5][i][j]][i] = true;
		}
	}
	// for degree 4
	higherHittingsVecs[4].resize(cellCount, boost::dynamic_bitset<> {higherDimUnav[4].size(), 0});
	for (unsigned int i = 0; i < higherDimUnav[4].size(); i++) {
		for (unsigned int j = 0; j < higherDimUnav[4][i].size(); j++) {
			higherHittingsVecs[4][higherDimUnav[4][i][j]][i] = true;
		}
	}
	// for degree 3
	higherHittingsVecs[3].resize(cellCount, boost::dynamic_bitset<> {higherDimUnav[3].size(), 0});
	for (unsigned int i = 0; i < higherDimUnav[3].size(); i++) {
		for (unsigned int j = 0; j < higherDimUnav[3][i].size(); j++) {
			higherHittingsVecs[3][higherDimUnav[3][i][j]][i] = true;
		}
	}
	// for degree 2
	higherHittingsVecs[2].resize(cellCount, boost::dynamic_bitset<> {higherDimUnav[2].size(), 0});
	for (unsigned int i = 0; i < higherDimUnav[2].size(); i++) {
		for (unsigned int j = 0; j < higherDimUnav[2][i].size(); j++) {
			higherHittingsVecs[2][higherDimUnav[2][i][j]][i] = true;
		}
	}
	return higherHittingsVecs;
}

// checks for a given grid whether there are any valid Sudoku with size <= n in it
bool doesValidSudokuExist(SudokuGrid& grid, vector<vector<byte> >& unavoidableSets, int n) {
	// sorting ua sets by size (in beforehand)
	sort(unavoidableSets.begin(), unavoidableSets.end(), 
		[](const vector<byte>& a, const vector<byte>& b) {
			return a.size() < b.size(); 
	});
	// for every unev set for every cell in it a vector with the cells to be dead if it is picked
	vector<vector<boost::dynamic_bitset<> > > deadclues(unavoidableSets.size());
	for (unsigned int i = 0; i < unavoidableSets.size(); i++) {
		deadclues[i].resize(unavoidableSets[i].size());
		for (unsigned int j = 0; j < unavoidableSets[i].size(); j++) {
			deadclues[i][j].resize((int)pow(grid.n,4));
			for (unsigned int k = 0; k < unavoidableSets[i].size(); k++) {
				int c = unavoidableSets[i][j];
				int d = unavoidableSets[i][k];
				if (d <= c) {
					deadclues[i][j][d] = 1;
				}
			}
		}
	}

	// for every picked cell during recursion save the cells which are dead already to improve performance
	vector<boost::dynamic_bitset<>> deadvec(n + 1, boost::dynamic_bitset<>{(int)pow(grid.n, 4),0});

	// for every unav set which cells are hit by it
	vector<boost::dynamic_bitset<>> hittingvec = initHittingVectors(grid, unavoidableSets);

	// which unavoid sets hit after k-th iteration
	vector<boost::dynamic_bitset<>> statevec(n + 1, boost::dynamic_bitset<>{unavoidableSets.size(), 0});

	// contains (incomplete) hitting sets for each iteration k
	vector<byte > hitset(n + 1, 0);

	size_t temp = getCurrentRSS();
	// create higher degree ua sets
	vector<vector<vector<byte>>> higherDimUnav = createHigherDimensionalUnavoidableSets(unavoidableSets, n);
	cout << "found " << higherDimUnav[2].size() << " ua sets of degree 2" << endl;
	cout << "found " << higherDimUnav[3].size() << " ua sets of degree 3" << endl;
	cout << "found " << higherDimUnav[4].size() << " ua sets of degree 4" << endl;
	cout << "found " << higherDimUnav[5].size() << " ua sets of degree 5" << endl;
	vector<vector<boost::dynamic_bitset<>>> higherHittingsVecs = createHigherHittingsVecs(higherDimUnav, (int)pow(grid.n, 4));
	vector<vector<boost::dynamic_bitset<>>> higherStateVec(6);
	// degree 5
	higherStateVec[5].resize(5, boost::dynamic_bitset<>{higherDimUnav[5].size(), 0});
	higherStateVec[5].resize(n + 1, boost::dynamic_bitset<>{1536, 0});
	// degree 4
	higherStateVec[4].resize(5, boost::dynamic_bitset<>{higherDimUnav[4].size(), 0});
	higherStateVec[4].resize(n + 1, boost::dynamic_bitset<>{1536, 0});
	// degree 3
	higherStateVec[3].resize(5, boost::dynamic_bitset<>{higherDimUnav[3].size(), 0});
	higherStateVec[3].resize(n + 1, boost::dynamic_bitset<>{1536, 0});
	// degree 2
	higherStateVec[2].resize(5, boost::dynamic_bitset<>{higherDimUnav[2].size(), 0});
	higherStateVec[2].resize(n + 1, boost::dynamic_bitset<>{1536, 0});

	// just for declaration
	vector<vector<boost::dynamic_bitset<>>> consolidatedHigherHittingVecs;
	consolidatedHigherHittingVecs.resize(6);
	// degree 5
	consolidatedHigherHittingVecs[5].resize(grid.n*grid.n*grid.n*grid.n, boost::dynamic_bitset<>{1536, 0});
	// degree 4
	consolidatedHigherHittingVecs[4].resize(grid.n*grid.n*grid.n*grid.n, boost::dynamic_bitset<>{1536, 0});
	// degree 3
	consolidatedHigherHittingVecs[3].resize(grid.n*grid.n*grid.n*grid.n, boost::dynamic_bitset<>{1536, 0});
	// degree 2
	consolidatedHigherHittingVecs[2].resize(grid.n*grid.n*grid.n*grid.n, boost::dynamic_bitset<>{1536, 0});
	cout << "memory just for higher degree sets: " << getCurrentRSS() - temp << endl;
	cout << "current memory usage before backtracking: " << getCurrentRSS() << endl;
	cout << "starts now" << endl;
	// run through backtracking, true is returned if there is a valid Sudoku below the size of n else false
	return addClue(0, n, 0, deadclues, deadvec, hittingvec, statevec, hitset, unavoidableSets,
		grid, higherHittingsVecs, higherStateVec, higherDimUnav, consolidatedHigherHittingVecs);
}

// Matlab equivalent
void tick() {
	c_start = clock();
	cycles_start = rdtsc();
}

// Matlab equivalent
void tock() {
	clock_t c_end = clock();
	uint64_t cycles_end = rdtsc();
		
	cout << "CPU time used: "
	<< 1000.0 * (c_end-c_start) / CLOCKS_PER_SEC << " ms\n" <<
	cycles_start << " " << cycles_end << " " << cycles_end - cycles_start << endl;
}

void tock(const string str) {
	clock_t c_end = clock();
	uint64_t cycles_end = rdtsc();

	cout << str << 1000.0 * (c_end - c_start) / CLOCKS_PER_SEC << " ms" << endl;
}




// accepts as parameter index of a 17-clue Sudoku in Gordon Royle's list and searches for valid 16-clue-Sudokus in it
// a reference to this list can be found in the file sudoku17-ml-credit.txt
int main(int argc, char* argv[]) {
	maxClues = 16;
	int index = 11;
	if (argc > 1) { // first parameter is program name
		index = atoi(argv[1]);
	}
	cout << "starts reading file, index is: " << index << endl;
	SudokuGrid grid = readSpecific17ClueSudoku(index);
	SudokuGridDomain domain(grid);
	tick();
	vector<SudokuGrid> solvedGrids = solveSudoku(grid, domain, false, true);
	tock("time to solve the " + std::to_string(index) + "-th 17-clue Sudoku: ");
	tick();
	vector<vector<byte> > unavoidableSets = findUnavoidableSets(solvedGrids[0]);
	cout << "still " << unavoidableSets.size() << " ua sets after duplicate detection" << endl;
	tock("time to find ua sets: ");
	tick();
	bool doesExist = doesValidSudokuExist(solvedGrids[0], unavoidableSets, maxClues);
	tock("time for whole search: ");
	cout << "time spend on checking candidates: " << 1000.0 * (timeSpentCheckingCandidates) / CLOCKS_PER_SEC << endl;
	cout << "valid " << maxClues << "-clue Sudoku does exist: " << doesExist << endl;
	cout << "searched through grid with index: " << index << endl;
	for (int i = 0; i < maxClues; i++) {
		cout << "count ignored " << countIgnored[i] << " in layer: " << i << endl;
		cout << "count not ignored " << countNotIgnored[i] << " in layer: " << i << endl;
	}
	for (int degree = 2; degree <= 5; degree++) {
		cout << "count backtracked " << countDegreeBacktracked[degree] << " for degree: " << degree << endl;
		cout << "count not backtracked " << countDegreeNotBacktracked[degree] << " for degree: " << degree << endl;
	}
	cout << "last layer and not hit all ua sets: " << countLastLayerNotAllHit << endl;
	cout << maxClues << "-clue hitting sets enumerated: " << countHittingSets << endl;
	cout << "peak memory usage: " << getPeakRSS() << endl;
	cout << "smaller hitting-sets: " << countSmallerHittingSets << endl;
	return 0;
}

// searches in i-th 4x4 Sudoku solution for a valid 3 clue Sudoku the Sudoku solutions are created in beforehand
int main2(int argc, char* argv[]) {
	maxClues = 3;
	int index = 4;
	if (argc > 1) { // first parameter is program name
		index = atoi(argv[1]);
	}
	// create all 4x4 solution grids (symmetric ones included)
	SudokuGrid grid(2);
	SudokuGridDomain domain(grid);
	vector<SudokuGrid> solvedGrids = solveSudoku(grid, domain, true, false);
	cout << "checking " << index << "-th 2x2 Sudoku solution grid" << endl;
	tick();
	vector<vector<byte> > unavoidableSets = findUnavoidableSets(solvedGrids[index]);
	cout << "still " << unavoidableSets.size() << " ua sets after duplicate detection" << endl;
	tock("time to find ua sets: ");
	tick();
	bool doesExist = doesValidSudokuExist(solvedGrids[index], unavoidableSets, maxClues);
	tock("time for whole search: ");
	cout << "time spend on checking candidates: " << 1000.0 * (timeSpentCheckingCandidates) / CLOCKS_PER_SEC << endl;
	cout << "valid " << maxClues << "-clue Sudoku does exist: " << doesExist << endl;
	cout << "searched through grid with index: " << index << endl;
	for (int i = 0; i < maxClues; i++) {
		cout << "count ignored " << countIgnored[i] << " in layer: " << i << endl;
		cout << "count not ignored " << countNotIgnored[i] << " in layer: " << i << endl;
	}
	for (int degree = 2; degree <= 5; degree++) {
		cout << "count backtracked " << countDegreeBacktracked[degree] << " for degree: " << degree << endl;
		cout << "count not backtracked " << countDegreeNotBacktracked[degree] << " for degree: " << degree << endl;
	}
	cout << "last layer and not hit all ua sets: " << countLastLayerNotAllHit << endl;
	cout << maxClues << "-clue hitting sets enumerated: " << countHittingSets << endl;
	cout << "peak memory usage: " << getPeakRSS() << endl;
	cout << "smaller hitting-sets: " << countSmallerHittingSets << endl;
	return 0;
}



